<?php

namespace Drupal\cas_server\Configuration;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\StringTranslation\TranslatableMarkup;

/**
 * Helper and utility class for retriving and checking values against config.
 */
class ConfigHelper {

  /**
   * Stores settings object.
   *
   * @var \Drupal\Core\Config\Config
   */
  protected $settings;

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * Constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory.
   * @param \Drupal\Core\Entity\EntityTypeManager $entity_manager
   *   The entity manager.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    EntityTypeManager $entity_manager
  ) {
    $this->settings = $config_factory->get('cas_server.settings');
    $this->entityTypeManager = $entity_manager;
  }

  /**
   * Check a service against the service whitelist.
   *
   * @param string $service
   *   A supplied service string.
   *
   * @return bool
   *   Whether or not the service is allowed.
   */
  public function checkServiceAgainstWhitelist($service) {
    return (bool) $this->matchServiceAgainstConfig($service);
  }

  /**
   * Return the list of attributes to be released for a service.
   *
   * @param string $service
   *   A supplied service string.
   *
   * @return array
   *   An array of user field names to be released as attributes.
   */
  public function getAttributesForService($service) {
    if ($def = $this->matchServiceAgainstConfig($service)) {
      return array_keys($def->getAttributes());
    }
    return [];
  }

  /**
   * Return the timeout for a proxy-granting ticket.
   *
   * @param int $now
   *   The expiry will be added to this to return expiry timestamp.
   *
   * @return int
   *   The number of seconds a proxy-granting ticket is valid.
   */
  public function getProxyGrantingTicketTimeout(int $now = -1) {
    $timeout = $this->settings->get('ticket.proxy_granting_ticket_timeout');
    if ($now > 0) {
      return $now + $timeout;
    }
    return $timeout;
  }

  /**
   * Return whether to use the ticket granting ticket or not.
   *
   * @return bool
   *   Whether to use the ticket granting cookie.
   */
  public function shouldUseTicketGrantingTicket() {
    return (bool) $this->settings->get('ticket.ticket_granting_ticket_auth');
  }

  /**
   * Return the timeout for a ticket-granting ticket.
   *
   * @param int $now
   *   The expiry will be added to this to return expiry timestamp.
   *
   * @return int
   *   The number of seconds a ticket-granting ticket is valid.
   */
  public function getTicketGrantingTicketTimeout(int $now = -1) {
    $timeout = $this->settings->get('ticket.ticket_granting_ticket_timeout');
    if ($now > 0) {
      return $now + $timeout;
    }
    return $timeout;
  }

  /**
   * Return the timeout for a proxy ticket.
   *
   * @param int $now
   *   The expiry will be added to this to return expiry timestamp.
   *
   * @return int
   *   The number of seconds a proxy ticket is valid.
   */
  public function getProxyTicketTimeout(int $now = -1) {
    $timeout = $this->settings->get('ticket.proxy_ticket_timeout');
    if ($now > 0) {
      return $now + $timeout;
    }
    return $timeout;
  }

  /**
   * Return the timeout for a service ticket.
   *
   * @return int
   *   The number of seconds a service ticket is valid.
   */
  public function getServiceTicketTimeout() {
    return $this->settings->get('ticket.service_ticket_timeout');
  }

  /**
   * The attribute to use for the username.
   *
   * @return string
   *   The username attribute.
   */
  public function getTicketUsernameAttribute() {
    $value = $this->settings->get('ticket.ticket_username_attribute');
    return $value ? $value : 'name';
  }

  /**
   * Return the custom not permitted message, or FALSE if not set.
   *
   * @return string|bool
   *   The configured service not permitted message.
   */
  public function getNotPermittedMessage() {
    if (!empty($m = $this->settings->get('messages.not_permitted'))) {
      return $m;
    }
    return FALSE;
  }

  /**
   * Check whether a service is configured for single sign on.
   *
   * @param string $service
   *   The service uri to check.
   *
   * @return bool
   *   Whether or not the service is authorized.
   */
  public function verifyServiceForSso($service) {
    if ($def = $this->matchServiceAgainstConfig($service)) {
      return $def->getSso();
    }
    return FALSE;
  }

  /**
   * Return the custom invalid service message, or FALSE.
   *
   * @return string|bool
   *   The configured invalid service message.
   */
  public function getInvalidServiceMessage() {
    if (!empty($m = $this->settings->get('messages.invalid_service'))) {
      return $m;
    }
    return FALSE;
  }

  /**
   * Return the custom user logout message, or FALSE.
   *
   * @return string|bool
   *   The configured user logout message.
   */
  public function getUserLogoutMessage() {
    if (!empty($m = $this->settings->get('messages.user_logout'))) {
      return $m;
    }
    return FALSE;
  }

  /**
   * Return the custom logged in message, or FALSE.
   *
   * @return string|bool
   *   The configured log in message.
   */
  public function getLoggedInMessage() {
    if (!empty($m = $this->settings->get('messages.logged_in'))) {
      return $m;
    }
    return FALSE;
  }

  /**
   * The attribute to use for authentication.
   *
   * @return string
   *   The username attribute.
   */
  public function getAuthenticationSourceField() {
    $value = $this->settings->get('login.username_attribute');
    return $value ? $value : 'name';
  }

  /**
   * The title of the login field on UserLogin form.
   *
   * @return \Drupal\Core\StringTranslation\TranslatableMarkup
   *   The translated title.
   */
  public function getAuthenticationSourceFieldTitle() {
    $title = NULL;

    $username_field = $this->getAuthenticationSourceField();
    switch ($username_field) {
      case 'both':
        $title = new TranslatableMarkup('Username or email address');
        break;

      case 'mail':
        $title = new TranslatableMarkup('Email address');
        break;

      default:
        $title = new TranslatableMarkup('Username');
    }

    return $title;
  }

  /**
   * Show reset password link on login form.
   *
   * @return bool
   *   Whether or not to show the reset password link.
   */
  public function getShowResetPassword() {
    $value = $this->settings->get('login.reset_password');
    return (bool) $value;
  }

  /**
   * Match a service string to a service definition.
   *
   * @param string $service
   *   The provided service string.
   *
   * @return CasServerServiceInterface|bool
   *   A matching CasServerService object or FALSE if no match.
   */
  public function matchServiceAgainstConfig($service) {
    if (empty($service)) {
      return FALSE;
    }

    $entity_manager = $this->entityTypeManager
      ->getStorage('cas_server_service');
    $sids = $entity_manager->getQuery()->accessCheck(FALSE)->execute();

    $service_definitions = $entity_manager->loadMultiple($sids);
    foreach ($service_definitions as $def) {
      if ($this->matchServiceString($def->getService(), $service)) {
        return $def;
      }
    }
    return FALSE;
  }

  /**
   * Match a string against a wildcard pattern.
   *
   * @param string $pattern
   *   The string pattern to match against.
   * @param string $service
   *   The string to try to match.
   *
   * @return bool
   *   Whether the string matched or not.
   *
   * @See https://www.drupal.org/project/config_ignore/issues/3182849
   */
  private function matchServiceString(string $pattern, string $service): bool {
    $pattern = '/^' . preg_quote($pattern, '/') . '$/';
    $pattern = str_replace('\*', '.*', $pattern);
    return (bool) preg_match($pattern, $service);
  }

}
