<?php

namespace Drupal\cas_server\Form;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;

/**
 * Configuration form for CAS Server settings.
 */
class CasServerSettings extends ConfigFormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'cas_server_settings';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config('cas_server.settings');

    $form['ticket'] = [
      '#type' => 'details',
      '#title' => $this->t('Ticket settings'),
      '#open' => TRUE,
      '#tree' => TRUE,
    ];
    $form['ticket']['service'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Service ticket timeout'),
      '#description' => $this->t('Time in seconds for which a service ticket is valid.'),
      '#size' => 30,
      '#default_value' => $config->get('ticket.service_ticket_timeout'),
    ];
    $form['ticket']['proxy'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Proxy ticket timeout'),
      '#description' => $this->t('Time in seconds for which a proxy ticket is valid.'),
      '#size' => 30,
      '#default_value' => $config->get('ticket.proxy_ticket_timeout'),
    ];
    $form['ticket']['proxy_granting'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Proxy granting ticket timeout'),
      '#description' => $this->t('Time in seconds for which a proxy granting ticket is valid.'),
      '#size' => 30,
      '#default_value' => $config->get('ticket.proxy_granting_ticket_timeout'),
    ];
    $form['ticket']['ticket_granting_auth'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use ticket granting ticket'),
      '#description' => $this->t('When checked, a user will be granted a ticket for login. If a user is logged in to Drupal, but does not have the ticket (or it has expired) then checking this will force them to enter their credentials again.'),
      '#size' => 30,
      '#default_value' => (bool) $config->get('ticket.ticket_granting_ticket_auth'),
    ];
    $form['ticket']['ticket_granting'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Ticket granting ticket timeout'),
      '#description' => $this->t('Time in seconds for which a ticket granting ticket is valid.'),
      '#size' => 30,
      '#default_value' => $config->get('ticket.ticket_granting_ticket_timeout'),
      '#states' => [
        'visible' => [
          ':input[name="ticket[ticket_granting_auth]"]' => ['checked' => TRUE],
        ],
      ],
    ];
    $form['ticket']['username'] = [
      '#type' => 'select',
      '#title' => $this->t('Username value'),
      '#description' => $this->t('Which value to use for the username to respond.'),
      '#options' => [
        'name' => $this->t('Username'),
        'mail' => $this->t('Email Address'),
        'uid' => $this->t('UID'),
      ],
      '#default_value' => $config->get('ticket.ticket_username_attribute'),
    ];

    $form['messages'] = [
      '#type' => 'details',
      '#title' => 'Custom Messages',
      '#open' => TRUE,
      '#tree' => TRUE,
    ];
    $form['messages']['invalid_service'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Display message for Invalid Service'),
      '#description' => $this->t('Message to display to a user requesting an invalid service.'),
      '#size' => 60,
      '#default_value' => $config->get('messages.invalid_service'),
    ];
    $form['messages']['not_permitted'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Display message for Not Permitted'),
      '#description' => $this->t("Message to display to a user requesting a service that they don't have permission to log in to."),
      '#size' => 60,
      '#default_value' => $config->get('messages.not_permitted'),
    ];
    $form['messages']['user_logout'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Display message for User Logout'),
      '#description' => $this->t('Message to display to a user logged out of single sign on.'),
      '#size' => 60,
      '#default_value' => $config->get('messages.user_logout'),
    ];
    $form['messages']['logged_in'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Display message for Logged In Users'),
      '#description' => $this->t('Message to display to a user already logged in to single sign on.'),
      '#size' => 60,
      '#default_value' => $config->get('messages.logged_in'),
    ];

    $form['debugging'] = [
      '#type' => 'details',
      '#title' => 'Debugging Options',
      '#open' => FALSE,
      '#tree' => TRUE,
    ];
    $form['debugging']['log'] = [
      '#type' => 'radios',
      '#title' => $this->t('Enable debugging log'),
      '#description' => $this->t('Enable debugging output to the Drupal log service.'),
      '#options' => [0 => 'No', 1 => 'Yes'],
      '#default_value' => $config->get('debugging.log'),
    ];

    $form['login'] = [
      '#type' => 'details',
      '#title' => 'Login Options',
      '#open' => FALSE,
      '#tree' => TRUE,
    ];
    $form['login']['username'] = [
      '#type' => 'select',
      '#title' => $this->t('Username field'),
      '#description' => $this->t('Which field to use for user authentication.'),
      '#options' => [
        'name' => $this->t('Username'),
        'mail' => $this->t('Email Address'),
        'both' => $this->t('Username or email address'),
      ],
      '#default_value' => $config->get('login.username_attribute'),
    ];
    $form['login']['reset_password'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Show reset password link on cas login form'),
      '#default_value' => $config->get('login.reset_password'),
    ];

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // Check timeouts to make sure they are integers.
    $values = $form_state->getValue('ticket');
    foreach ($values as $key => $value) {
      if ($key == 'username') {
        continue;
      }
      if (!is_numeric($value) || !(round($value) == $value)) {
        $form_state->setErrorByName(
          "ticket][$key",
          $this->t('Ticket timeouts must be integer-valued')
        );
      }
    }

    return parent::validateForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $config = $this->config('cas_server.settings');
    $ticket_data = $form_state->getValue('ticket');
    $config
      ->set('ticket.service_ticket_timeout', (int) $ticket_data['service'])
      ->set('ticket.proxy_ticket_timeout', (int) $ticket_data['proxy'])
      ->set('ticket.proxy_granting_ticket_timeout', (int) $ticket_data['proxy_granting'])
      ->set('ticket.ticket_granting_ticket_auth', (int) $ticket_data['ticket_granting_auth'])
      ->set('ticket.ticket_granting_ticket_timeout', (int) $ticket_data['ticket_granting'])
      ->set('ticket.ticket_username_attribute', $ticket_data['username']);

    $message_data = $form_state->getValue('messages');
    $config
      ->set('messages.invalid_service', $message_data['invalid_service'])
      ->set('messages.not_permitted', $message_data['not_permitted'])
      ->set('messages.user_logout', $message_data['user_logout'])
      ->set('messages.logged_in', $message_data['logged_in']);

    $config->set('debugging.log', (bool) $form_state->getValue(['debugging', 'log']));
    $config->set('login.username_attribute', $form_state->getValue(['login', 'username']));
    $config->set('login.reset_password', $form_state->getValue(['login', 'reset_password']));
    $config->save();

    parent::submitForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['cas_server.settings'];
  }

}
