<?php

namespace Drupal\cas_server\Ticket;

use Drupal\cas_server\Configuration\ConfigHelper;
use Drupal\cas_server\Event\CasServerTicketAlterEvent;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Utility\Crypt;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\SessionManagerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Build appropriate Tickets based on inputs.
 */
class TicketFactory {

  /**
   * The ticket store.
   *
   * @var \Drupal\cas_server\Ticket\TicketStorageInterface
   */
  protected $ticketStore;

  /**
   * The configuration helper.
   *
   * @var \Drupal\cas_server\Configuration\ConfigHelper
   */
  protected $configHelper;

  /**
   * The session manager.
   *
   * @var \Drupal\Core\Session\SessionManagerInterface
   */
  protected $sessionManager;

  /**
   * Event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Constructor.
   *
   * @param \Drupal\cas_server\Ticket\TicketStorageInterface $ticket_store
   *   The ticket store to use.
   * @param \Drupal\cas_server\Configuration\ConfigHelper $config_helper
   *   The configuration helper.
   * @param \Drupal\Core\Session\SessionManagerInterface $session_manager
   *   The session manager.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\Component\Datetime\TimeInterface $time_service
   *   The time service.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   */
  public function __construct(
    TicketStorageInterface $ticket_store,
    ConfigHelper $config_helper,
    SessionManagerInterface $session_manager,
    EventDispatcherInterface $event_dispatcher,
    TimeInterface $time_service,
    AccountInterface $current_user
  ) {
    $this->ticketStore = $ticket_store;
    $this->configHelper = $config_helper;
    $this->sessionManager = $session_manager;
    $this->eventDispatcher = $event_dispatcher;
    $this->time = $time_service;
    // @todo Determine if injecting this means we have the wrong user.
    $this->currentUser = $current_user;
  }

  /**
   * Create a proxy granting ticket.
   *
   * @param array $proxy_chain
   *   The proxy chain.
   *
   * @return ProxyGrantingTicket
   *   The created and saved PGT.
   */
  public function createProxyGrantingTicket($proxy_chain) {
    $id = 'PGT-';
    $id .= Crypt::randomBytesBase64(32);
    $expiration_time = $this->time->getRequestTime()
      + $this->configHelper->getProxyGrantingTicketTimeout();
    $session = Crypt::hashBase64($this->sessionManager->getId());
    $uid = $this->currentUser->id();
    $name = $this->getUsernameAttribute($this->currentUser);

    $pgt = new ProxyGrantingTicket(
      $id,
      $expiration_time,
      $session,
      $uid,
      $name,
      $proxy_chain
    );
    $event = new CasServerTicketAlterEvent($pgt);
    $this->eventDispatcher->dispatch(
      $event,
      CasServerTicketAlterEvent::CAS_SERVER_TICKET_ALTER_EVENT
    );
    $this->ticketStore->storeProxyGrantingTicket($pgt);

    return $pgt;
  }

  /**
   * Create a service ticket.
   *
   * @param string $service_string
   *   The service the ticket is being issued for.
   * @param bool $renew
   *   If this ticket was generated by directly supplied credentials.
   *
   * @return ServiceTicket
   *   The created and saved ServiceTicket.
   */
  public function createServiceTicket($service_string, $renew) {
    $id = 'ST-';
    $id .= Crypt::randomBytesBase64(32);
    $expiration_time = $this->time->getRequestTime()
      + $this->configHelper->getServiceTicketTimeout();
    $session = Crypt::hashBase64($this->sessionManager->getId());
    $uid = $this->currentUser->id();
    $name = $this->getUsernameAttribute($this->currentUser);

    $st = new ServiceTicket(
      $id,
      $expiration_time,
      $session,
      $uid,
      $name,
      $service_string,
      $renew
    );
    $event = new CasServerTicketAlterEvent($st);
    $this->eventDispatcher->dispatch(
      $event,
      CasServerTicketAlterEvent::CAS_SERVER_TICKET_ALTER_EVENT
    );
    $this->ticketStore->storeServiceTicket($st);

    return $st;
  }

  /**
   * Create a proxy ticket.
   *
   * @param string $service_string
   *   The service the ticket is being issued for.
   * @param bool $renew
   *   If this ticket was generated by directly supplied credentials.
   * @param array $proxy_chain
   *   The proxy chain.
   * @param string $session
   *   The ID of the session who originally requested the requesting pgt.
   * @param int $uid
   *   The uid of the of user who originally requested the pgt.
   * @param string $name
   *   The username of the user who originally requested the requesting pgt.
   *
   * @return ProxyTicket
   *   The created and saved ProxyTicket.
   */
  public function createProxyTicket(
    $service_string,
    $renew,
    $proxy_chain,
    $session,
    $uid,
    $name
  ) {
    $id = 'PT-';
    $id .= Crypt::randomBytesBase64(32);
    $expiration_time = $this->time->getRequestTime()
      + $this->configHelper->getProxyTicketTimeout();

    $pt = new ProxyTicket(
      $id,
      $expiration_time,
      $session,
      $uid,
      $name,
      $service_string,
      $renew,
      $proxy_chain
    );
    $event = new CasServerTicketAlterEvent($pt);
    $this->eventDispatcher->dispatch(
      $event,
      CasServerTicketAlterEvent::CAS_SERVER_TICKET_ALTER_EVENT
    );
    $this->ticketStore->storeProxyTicket($pt);

    return $pt;
  }

  /**
   * Create a ticket granting ticket.
   *
   * @return TicketGrantingTicket
   *   The created and saved TicketGrantingTicket.
   */
  public function createTicketGrantingTicket() {
    $id = 'TGT-';
    $id .= Crypt::randomBytesBase64(32);
    $expiration_time = $this->time->getRequestTime()
      + $this->configHelper->getTicketGrantingTicketTimeout();
    $session = Crypt::hashBase64($this->sessionManager->getId());
    $uid = $this->currentUser->id();
    $name = $this->getUsernameAttribute($this->currentUser);

    $tgt = new TicketGrantingTicket(
      $id,
      $expiration_time,
      $session,
      $uid,
      $name
    );
    $event = new CasServerTicketAlterEvent($tgt);
    $this->eventDispatcher->dispatch(
      $event,
      CasServerTicketAlterEvent::CAS_SERVER_TICKET_ALTER_EVENT
    );
    $this->ticketStore->storeTicketGrantingTicket($tgt);

    return $tgt;
  }

  /**
   * Get the username for a ticket.
   *
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The account to get the username for.
   *
   * @return string
   *   The username.
   */
  public function getUsernameAttribute(AccountInterface $account) {
    if ($this->configHelper->getTicketUsernameAttribute() == 'mail') {
      $name = $account->getEmail();
    }
    elseif ($this->configHelper->getTicketUsernameAttribute() == 'uid') {
      $name = $account->id();
    }
    else {
      $name = $account->getAccountName();
    }

    return $name;
  }

}
