<?php

namespace Drupal\Tests\cas_server\Functional;

use Drupal\cas_server\Entity\CasServerService;
use Drupal\Tests\BrowserTestBase;

/**
 * Tests responses from the proxy ticket granting system.
 *
 * @group cas_server
 */
class ProxyControllerTest extends BrowserTestBase {

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'cas_server',
  ];

  /**
   * An user with Anonymous permissions.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $exampleUser;

  /**
   * The ticket factory.
   *
   * @var \Drupal\cas_server\Ticket\TicketFactory
   */
  protected $ticketFactory;

  /**
   * The ticket store.
   *
   * @var \Drupal\cas_server\Ticket\TicketStorageInterface
   */
  protected $ticketStore;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->exampleUser = $this->drupalCreateUser([], 'exampleUserName');

    $this->ticketFactory = $this->container->get('cas_server.ticket_factory');
    $this->ticketStore = $this->container->get('cas_server.storage');

    $test_proxy = CasServerService::create([
      'id' => 'test_proxy',
      'label' => 'Proxyable Test Service',
      'service' => 'https://foo.example.com*',
      'sso' => TRUE,
      'attributes' => [],
    ]);
    $test_proxy->save();

    $test_no_proxy = CasServerService::create([
      'id' => 'test_no_proxy',
      'label' => 'Unproxyable Test Service',
      'service' => 'https://bar.example.com*',
      'sso' => FALSE,
      'attributes' => [],
    ]);
    $test_no_proxy->save();
  }

  /**
   * Tests an successful request.
   */
  public function testProxySuccessRequest() {
    $this->drupalLogin($this->exampleUser);
    $pgt = $this->ticketFactory->createProxyGrantingTicket([]);
    $this->drupalLogout();
    $service = 'https://foo.example.com';

    $this->drupalGet('cas/proxy', [
      'query' => [
        'pgt' => $pgt->getId(),
        'targetService' => $service,
      ],
    ]);
    $this->assertSession()->responseContains("<cas:proxySuccess>");
    $this->assertSession()->responseContains("<cas:proxyTicket>");
    $this->assertSession()->statusCodeEquals(200);
  }

  /**
   * Tests an invalid proxy request.
   */
  public function testInvalidProxyRequest() {
    $this->drupalGet('cas/proxy');
    $this->assertSession()
      ->responseContains("<cas:proxyFailure code='INVALID_REQUEST'>");
    $this->assertSession()->statusCodeEquals(200);
  }

  /**
   * Tests an unauthorized service request.
   */
  public function testUnauthorizedServiceRequest() {
    $this->drupalLogin($this->exampleUser);
    $pgt = $this->ticketFactory->createProxyGrantingTicket([]);
    $this->drupalLogout();
    $service = 'https://bar.example.com';

    $this->drupalGet('cas/proxy', [
      'query' => [
        'pgt' => $pgt->getId(),
        'targetService' => $service,
      ],
    ]);
    $this->assertSession()
      ->responseContains("<cas:proxyFailure code='UNAUTHORIZED_SERVICE'>");
    $this->assertSession()->statusCodeEquals(200);
  }

  /**
   * Tests an expired ticket request.
   */
  public function testExpiredTicketRequest() {
    $this->config('cas_server.settings')
      ->set('ticket.proxy_granting_ticket_timeout', -20)
      ->save();
    $this->drupalLogin($this->exampleUser);
    $pgt = $this->ticketFactory->createProxyGrantingTicket([]);
    $this->drupalLogout();
    $service = 'https://foo.example.com';

    $this->drupalGet('cas/proxy', [
      'query' => [
        'pgt' => $pgt->getId(),
        'targetService' => $service,
      ],
    ]);
    $this->assertSession()
      ->responseContains("<cas:proxyFailure code='INVALID_TICKET'>");
    $this->assertSession()->statusCodeEquals(200);
  }

  /**
   * Tests a missing ticket request.
   */
  public function testMissingTicketRequest() {
    $this->drupalLogin($this->exampleUser);
    $pgt = $this->ticketFactory->createProxyGrantingTicket([]);
    $this->drupalLogout();
    $this->ticketStore->deleteProxyGrantingTicket($pgt);
    $service = 'https://foo.example.com';

    $this->drupalGet('cas/proxy', [
      'query' => [
        'pgt' => $pgt->getId(),
        'targetService' => $service,
      ],
    ]);
    $this->assertSession()
      ->responseContains("<cas:proxyFailure code='INVALID_REQUEST'>");
    $this->assertSession()->pageTextContains("Ticket not found");
    $this->assertSession()->statusCodeEquals(200);
  }

  /**
   * Tests a incorrect ticket type request.
   */
  public function testWrongTicketTypeRequest() {
    $this->drupalLogin($this->exampleUser);
    $pgt = $this->ticketFactory->createTicketGrantingTicket();
    $this->drupalLogout();
    $service = 'https://foo.example.com';

    $this->drupalGet('cas/proxy', [
      'query' => [
        'pgt' => $pgt->getId(),
        'targetService' => $service,
      ],
    ]);
    $this->assertSession()
      ->responseContains("<cas:proxyFailure code='INVALID_REQUEST'>");
    $this->assertSession()->pageTextNotContains("Ticket not found");
    $this->assertSession()->statusCodeEquals(200);
  }

}
