<?php

namespace Drupal\Tests\cas_server\Functional;

use Drupal\cas_server\Entity\CasServerService;
use Drupal\Component\Utility\Crypt;
use Drupal\Core\Url;
use Drupal\Tests\BrowserTestBase;

/**
 * Tests responses from the user login form.
 *
 * @group cas_server
 */
class UserLoginFormTest extends BrowserTestBase {

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'cas_server',
  ];

  /**
   * Configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The ticket factory.
   *
   * @var \Drupal\cas_server\Ticket\TicketFactory
   */
  protected $ticketFactory;

  /**
   * The ticket store.
   *
   * @var \Drupal\cas_server\Ticket\TicketStorageInterface
   */
  protected $ticketStore;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * Drupal\Core\TempStore\PrivateTempStoreFactory definition.
   *
   * @var \Drupal\Core\TempStore\PrivateTempStoreFactory
   */
  protected $tempStoreFactory;

  /**
   * CAS Service Service entity storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $entityStorage;

  /**
   * A user with Anonymous permissions, but can log into any service.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $exampleUser;

  /**
   * A user with permission to log into test service.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $userCanAccess;

  /**
   * A user with Anonymous permission. Cannot access services.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $userCannotAccess;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->configFactory = $this->container->get('config.factory');
    $this->ticketFactory = $this->container->get('cas_server.ticket_factory');
    $this->ticketStore = $this->container->get('cas_server.storage');
    $this->connection = $this->container->get('database');
    $this->tempStoreFactory = $this->container->get('tempstore.private');
    $this->entityStorage = $this->container->get('entity_type.manager')
      ->getStorage('cas_server_service');

    $test = CasServerService::create([
      'id' => 'test',
      'label' => 'Test Service',
      'service' => '*',
      'sso' => TRUE,
      'attributes' => [],
    ]);
    $test->save();

    $this->exampleUser = $this->drupalCreateUser(
      ['cas server login to any service'],
      'exampleUserName'
    );

    $this->userCanAccess = $this->drupalCreateUser(
      ['cas server login to test service'],
      'userCanAccess'
    );

    $this->userCannotAccess = $this->drupalCreateUser(
      [],
      'userCannotAccess'
    );

    // Remove default permission added during install so that permissions can
    // be tested with existing tests.
    user_role_revoke_permissions(
      'authenticated', [
        'cas server login to any service',
      ]
    );
  }

  /**
   * Test submitting with bad username/password.
   */
  public function testBadCredentials() {
    $this->drupalGet('cas/login');
    $edit = [
      'username' => $this->exampleUser->getAccountName(),
      'password' => $this->exampleUser->pass_raw . 'foadasd',
    ];
    $this->submitForm($edit, 'Submit');

    $cookie_cas_tgc = $this->getSession()->getCookie('cas_tgc');
    $this->assertEmpty($cookie_cas_tgc);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()
      ->pageTextContains('Invalid username or password. Please try again.');
  }

  /**
   * Test submitting with correct values but no service, and no TGC enabled.
   */
  public function testCorrectNoServiceNoTicketGrantingCookie() {
    $this->drupalGet('cas/login');
    $edit = [
      'username' => $this->exampleUser->getAccountName(),
      'password' => $this->exampleUser->pass_raw,
    ];
    $this->submitForm($edit, 'Submit');

    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()
      ->pageTextContains('You are logged in to CAS single sign on.');

    // No cookie set because ticket_granting_ticket_auth defaults to false.
    $cookie_cas_tgc = $this->getSession()->getCookie('cas_tgc');
    $this->assertEmpty($cookie_cas_tgc);
  }

  /**
   * Test submitting with correct values but no service, with TGC enabled.
   */
  public function testCorrectNoServiceWithTicketGrantingCookie() {
    // Change the configuration setting.
    $editable = $this->configFactory->getEditable('cas_server.settings');
    $editable->set('ticket.ticket_granting_ticket_auth', TRUE);
    $editable->save();

    $this->drupalGet('cas/login');
    $edit = [
      'username' => $this->exampleUser->getAccountName(),
      'password' => $this->exampleUser->pass_raw,
    ];
    $this->submitForm($edit, 'Submit');

    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()
      ->pageTextContains('You are logged in to CAS single sign on.');

    $cookie_cas_tgc = $this->getSession()->getCookie('cas_tgc');
    $this->assertNotEmpty($cookie_cas_tgc);
  }

  /**
   * Test submitting without a valid login ticket.
   */
  public function testInvalidLoginTicket() {
    // We want to cause the session lt to no longer be valid; since that is
    // stored in the session, it must be destroyed.
    $this->drupalGet('cas/login');
    $this->getSession()->reset();
    $edit = [
      'username' => $this->exampleUser->getAccountName(),
      'password' => $this->exampleUser->pass_raw,
    ];
    $this->submitForm($edit, 'Submit');

    $this->assertSession()->statusCodeEquals(200);
    $cookie_cas_tgc = $this->getSession()->getCookie('cas_tgc');
    $this->assertEmpty($cookie_cas_tgc);
    $this->assertSession()
      ->pageTextContains('Login ticket invalid. Please try again');
  }

  /**
   * Test submitting with correct values and a service.
   */
  public function testCorrectWithService() {
    // Change the configuration setting to allow the possibility of TGC.
    $editable = $this->configFactory->getEditable('cas_server.settings');
    $editable->set('ticket.ticket_granting_ticket_auth', TRUE);
    $editable->save();

    $service = Url::fromRoute('cas_server.validate1');
    $service->setAbsolute();
    $this->drupalGet('cas/login', [
      'query' => [
        'service' => $service->toString(),
      ],
    ]);
    $edit = [
      'username' => $this->exampleUser->getAccountName(),
      'password' => $this->exampleUser->pass_raw,
    ];
    $this->submitForm($edit, 'Submit');

    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()
      ->pageTextNotContains('Your account does not have the required permissions to log in to this service');

    $cookie_cas_tgc = $this->getSession()->getCookie('cas_tgc');
    $this->assertNotEmpty($cookie_cas_tgc);

    // @phpcs:disable
    // @todo Left-over from WebTestBase::redirectCount. Find alternative.
    // $this->assertEquals($this->redirectCount, 2);
    // @phpcs:enable

    $session_id = $this
      ->getSession()
      ->getCookie($this->getSessionName());
    $ticket = $this->connection->select('cas_server_ticket_store', 'c')
      ->fields('c', ['id'])
      ->condition('session', Crypt::hashBase64($session_id))
      ->condition('type', 'service')
      ->execute()
      ->fetch();
    $this->assertNotEmpty($ticket);
    $tid = $ticket->id;

    // Validate the ticket and confirm the url remains the same.
    $post_login_url = $this->getSession()->getCurrentUrl();
    $this->drupalGet('cas/validate', [
      'query' => [
        'ticket' => $tid,
      ],
    ]);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->addressEquals($post_login_url);
  }

  /**
   * Test submitting with correct values and a service not configured for SSO.
   */
  public function testCorrectWithServiceNoSso() {
    // Change the configuration setting to allow the possibility of TGC.
    $editable = $this->configFactory->getEditable('cas_server.settings');
    $editable->set('ticket.ticket_granting_ticket_auth', TRUE);
    $editable->save();

    $test = $this->entityStorage->load('test');
    $test->setSso(FALSE);
    $test->save();

    $service = Url::fromRoute('cas_server.validate1');
    $service->setAbsolute();
    $this->drupalGet('cas/login', [
      'query' => [
        'service' => $service->toString(),
      ],
    ]);
    $edit = [
      'username' => $this->exampleUser->getAccountName(),
      'password' => $this->exampleUser->pass_raw,
    ];
    $this->submitForm($edit, 'Submit');

    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()
      ->pageTextNotContains('Your account does not have the required permissions to log in to this service');

    $cookie_cas_tgc = $this->getSession()->getCookie('cas_tgc');
    $this->assertEmpty($cookie_cas_tgc);

    // @phpcs:disable
    // @todo Left-over from WebTestBase::redirectCount. Find alternative.
    // $this->assertEquals($this->redirectCount, 2);
    // @phpcs:enable

    $session_id = $this
      ->getSession()
      ->getCookie($this->getSessionName());
    $ticket = $this->connection->select('cas_server_ticket_store', 'c')
      ->fields('c', ['id'])
      ->condition('session', Crypt::hashBase64($session_id))
      ->condition('type', 'service')
      ->execute()
      ->fetch();
    $this->assertNotEmpty($ticket);
    $tid = $ticket->id;

    // Validate the ticket and confirm the url remains the same.
    $post_login_url = $this->getSession()->getCurrentUrl();
    $this->drupalGet('cas/validate', [
      'query' => [
        'ticket' => $tid,
      ],
    ]);
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->addressEquals($post_login_url);
  }

  /**
   * Test submitting with correct values and a service.
   */
  public function testServicePermissions() {

    // Change the configuration setting to allow the possibility of TGC.
    $editable = $this->configFactory->getEditable('cas_server.settings');
    $editable->set('ticket.ticket_granting_ticket_auth', TRUE);
    $editable->save();

    // Use validate url as service.
    $service = Url::fromRoute('cas_server.validate1');
    $service->setAbsolute();
    $this->drupalGet('cas/login', [
      'query' => [
        'service' => $service->toString(),
      ],
    ]);

    // User cannot access given service.
    $edit = [
      'username' => $this->userCannotAccess->getAccountName(),
      'password' => $this->userCannotAccess->pass_raw,
    ];
    $this->submitForm($edit, 'Submit');

    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()
      ->pageTextContains('Your account does not have the required permissions to log in to this service');

    $cookie_cas_tgc = $this->getSession()->getCookie('cas_tgc');
    $this->assertEmpty($cookie_cas_tgc);

    // User can access given service. Call cas/login for new LT.
    $this->drupalGet('cas/login', [
      'query' => [
        'service' => $service->toString(),
      ],
    ]);
    $edit = [
      'username' => $this->userCanAccess->getAccountName(),
      'password' => $this->userCanAccess->pass_raw,
    ];
    $this->submitForm($edit, 'Submit');

    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()
      ->pageTextNotContains('Your account does not have the required permissions to log in to this service');

    $cookie_cas_tgc = $this->getSession()->getCookie('cas_tgc');
    $this->assertNotEmpty($cookie_cas_tgc);
  }

}
