<?php

namespace Drupal\Tests\cas_server\Unit\Controller;

use Drupal\cas_server\Ticket\ProxyTicket;
use Drupal\Tests\cas_server\Unit\TestTicketValidationController;
use Drupal\Tests\UnitTestCase;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\TransferException;
use GuzzleHttp\Handler\MockHandler;
use GuzzleHttp\HandlerStack;
use GuzzleHttp\Psr7\Response;

/**
 * TicketValidationController unit tests.
 *
 * @group cas_server
 *
 * @coversDefaultClass \Drupal\cas_server\Controller\TicketValidationController
 */
class TicketValidationControllerTest extends UnitTestCase {

  /**
   * The ticket factory.
   *
   * @var \Drupal\cas_server\Ticket\TicketFactory
   */
  protected $ticketFactory;

  /**
   * A proxy granting ticket.
   *
   * @var \Drupal\cas_server\Ticket\ProxyGrantingTicket
   */
  protected $pgt;

  /**
   * The ticket storage.
   *
   * @var \Drupal\cas_server\Ticket\TicketStorageInterface
   */
  protected $ticketStore;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->ticketFactory = $this
      ->createMock('Drupal\cas_server\Ticket\TicketFactory');

    $this->pgt = $this
      ->createMock('Drupal\cas_server\Ticket\ProxyGrantingTicket');

    $this->ticketStore = $this
      ->createMock('Drupal\cas_server\Ticket\TicketStorageInterface');
  }

  /**
   * Test the proxyCallback method through the test class.
   *
   * @covers ::proxyCallback
   *
   * @dataProvider proxyCallbackSuccessDataProvider
   */
  public function testProxyCallbackSuccess($url, $ticket) {
    $mock = new MockHandler([
      new Response(200),
      new Response(200),
    ]);
    $handler = HandlerStack::create($mock);
    $client = new Client(['handler' => $handler]);

    $this->ticketFactory->expects($this->once())
      ->method('createProxyGrantingTicket')
      ->will($this->returnValue($this->pgt));

    $this->pgt->expects($this->once())
      ->method('getId')
      ->will($this->returnValue('thisisapgtid'));

    // Only mock the getProxyChain for ProxyTicket.
    if ($ticket instanceof ProxyTicket) {
      $ticket->expects($this->any())
        ->method('getProxyChain')
        ->will($this->returnValue(['foo']));
    }

    $controller = new TestTicketValidationController(
      $client,
      $this->ticketFactory,
      $this->ticketStore
    );
    $this->assertNotFalse($controller->callProxyCallback($url, $ticket));
  }

  /**
   * Data provider for testProxyCallbackSuccess.
   */
  public function proxyCallbackSuccessDataProvider() {

    $urls = ['https://example.com', 'https://example.com/bar?q=foo'];
    $service_ticket = $this->createMock('Drupal\cas_server\Ticket\ServiceTicket');
    $proxy_ticket = $this->createMock('Drupal\cas_server\Ticket\ProxyTicket');

    return [
      [$urls[0], $service_ticket],
      [$urls[1], $service_ticket],
      [$urls[0], $proxy_ticket],
      [$urls[1], $proxy_ticket],
    ];
  }

  /**
   * Test failure conditions for proxyCallback.
   *
   * @covers ::proxyCallback
   *
   * @dataProvider proxyCallbackFailureDataProvider
   */
  public function testProxyCallbackFailure($url, $client) {

    $this->ticketFactory->expects($this->any())
      ->method('createProxyGrantingTicket')
      ->will($this->returnValue($this->pgt));

    $this->pgt->expects($this->any())
      ->method('getId')
      ->will($this->returnValue('thisisapgtid'));

    $st = $this->createMock('Drupal\cas_server\Ticket\ServiceTicket');

    $controller = new TestTicketValidationController(
      $client,
      $this->ticketFactory,
      $this->ticketStore
    );
    $this->assertFalse($controller->callProxyCallback($url, $st));

  }

  /**
   * Data provider for testProxyCallbackFailure.
   */
  public function proxyCallbackFailureDataProvider() {
    $urls = ['http://example.com', 'https://example.com'];

    $mock = new MockHandler([
      new TransferException(),
      new Response(200),
    ]);
    $handler = HandlerStack::create($mock);
    $client1 = new Client(['handler' => $handler]);

    $mock = new MockHandler([
      new Response(200),
      new TransferException(),
    ]);
    $handler = HandlerStack::create($mock);
    $client2 = new Client(['handler' => $handler]);

    return [
      [$urls[0], $client1],
      [$urls[1], $client2],
      [$urls[1], $client1],
    ];
  }

}
